#!/usr/bin/perl -w
use strict;
use FileHandle;
use IPC::Open2;
use POSIX ":sys_wait_h";

# this shell potentially allows you to connect any classifier to any
# search algorithm.  it does so through the use of pipes, so there's
# no compilation/linking step.  the caveat is that this makes the
# algorithm (somewhat) needlessly slow :).  i would recommend using it
# only for prototyping.
#
# in order to use SearnShell, you need to provide a single
# configuration file, specifying both the search problem and the
# learner.  an example config file for sequence labeling using
# multiclass megam as the classifier is given in sequence.config.

my %ALLOWEDVALUES = ();
$ALLOWEDVALUES{'costs'}{'optimal'}               = 1;   # only supporting optimal completion
$ALLOWEDVALUES{'costs'}{'simulate'}              = 1;   # only supporting optimal completion
$ALLOWEDVALUES{'actions'}{'fixed'}               = 1;   # same actions supported for each step
$ALLOWEDVALUES{'actions'}{'variable'}            = 1;   # variable action set
$ALLOWEDVALUES{'classifier_bzip'}{'true'}        = 1; 
$ALLOWEDVALUES{'classifier_bzip'}{'false'}       = 1; 
$ALLOWEDVALUES{'classifier_type'}{'multiclass'}  = 1;
$ALLOWEDVALUES{'classifier_type'}{'multilabel'}  = 1;
$ALLOWEDVALUES{'classifier_type'}{'singletons'}  = 1;
$ALLOWEDVALUES{'classifier_usedev'}{'true'}      = 1;   # include "DEV" data ala megam
$ALLOWEDVALUES{'classifier_usedev'}{'false'}     = 1;   # include "DEV" data ala megam
$ALLOWEDVALUES{'classifier_features'}{'numeric'} = 1;   # features must be numeric 1...N
$ALLOWEDVALUES{'classifier_features'}{'text'}    = 1;   # features can be anything
$ALLOWEDVALUES{'classifier_features'}{'svm'}     = 1;   # features must be #:# format (ala svmlight/libsvm)

my $configFile = shift or die "usage: SearnShell.pl <config file> <training data> <dev data>\n";
my $trainFile  = shift or die "usage: SearnShell.pl <config file> <training data> <dev data>\n";
my $testFile   = shift or die "usage: SearnShell.pl <config file> <training data> <dev data>\n";

my %cfg = load_config($configFile);

# DICTIONARY FOR CLASSES
my %classes = (); my %classesR = (); my $classesN = 0;

# DICTIONARY FOR FEATURES
my %features = (); my $featuresN = 1;

$|++;
$SIG{CHLD} = \&REAPER;

# initialize the policy
my %policy = ();
$policy{0}{WEIGHT} = 1;
$policy{0}{TYPE}   = 'optimal';

my $tempnumber = int(rand()*10000000000000);

# load the search
my $searchPid = open2(\*SearchR, \*SearchW, $cfg{'search_exec'})
    or die ("cannot open bidirectional pipe for executable " . $cfg{'search_exec'});

# wait for "ok"
my $waitForOkay = <SearchR>; chomp $waitForOkay;
if ($waitForOkay ne 'OK') { die "executable " . $cfg{'search_exec'} . " did not return an initial 'OK'"; }

for (my $iteration=1; $iteration<=4; $iteration++) {
    print STDERR "ITERATION $iteration\n";

    # begin writing examples
    my $TMPDATA = ".tmp.$tempnumber.data.$iteration";
    my @outputs = ();
    my $XV = $cfg{'xv'}; if ($XV eq 'none') { $XV = 1; }
    for (my $xv=0; $xv<$XV; $xv++) {
        local *XVD;
        if ($cfg{'classifier_bzip'} eq 'true') {
            open XVD, "| bzip2 -9 > $TMPDATA.$xv.bz2" or die "could not open $TMPDATA.$xv.bz2 for writing\n";
        } else {
            open XVD, "> $TMPDATA.$xv" or die "could not open $TMPDATA.$xv for writing\n";
        }
        $outputs[$xv] = *XVD{IO};
    }
    if ($XV > 1) {
        if ($cfg{'classifier_bzip'} eq 'true') {
            open D, "| bzip2 -9 > $TMPDATA.bz2" or die "could not open $TMPDATA.bz2 for writing\n";
        } else {
            open D, "> $TMPDATA.bz2" or die "could not open $TMPDATA.bz2 for writing\n";
        }
    }

    if ($cfg{'active_iteration'} <= $iteration) {
        print STDERR "active pass...";
        issueCommand("load " . $cfg{'active_data'});
        my $num = issueCommand("num");
        my $count = int($num/50); if ($count <= 0) { $count = 1; }
        for (my $n=0; $n<$num; $n++) {
            if (($n % $count) == 0) { print STDERR "."; }
            # first, get the true sequence
            my @origStates = ();
            my @origActions = ();
            my @origSeconds = ();
            my @origUncert  = ();
            my $s = issueCommand("example $n");
            my $numSteps = 0;
            while (1) {
                my @actions  = split /[\s]+/, issueCommand("actions $s");
                if (@actions == 0) { last; }  # final state

                my @features;
                if ($cfg{'actions'} eq 'fixed') {
                    @features = split /[\s]+/, issueCommand("features $s");
                } else {
                    for (my $i=0; $i<@actions; $i++) {
                        @{$features[$i]} = split /[\s]+/, issueCommand("action_features $s $actions[$i]");
                    }
                }
                my $pol = select_policy(0);
                my ($p,@probs) = getPrediction($pol, ($XV==1)?0:(-1), \@actions, \@features, 1);
                my ($u,$ustep) = uncertainty(@probs);

#                $p = getPrediction(0, ($XV==1)?0:(-1), \@actions, \@features, 1); # optimal

                push @origStates , $s;
                push @origActions, $p;
                push @origSeconds, $actions[$ustep];
                push @origUncert , $u;

                $s = issueCommand("step $s $p");
                $numSteps++;
            }

            my $standardLoss = issueCommand("loss $s");

            for (my $oi=0; $oi<@origStates; $oi++) {
                my $p = $origActions[$oi];
                my $u = $origSeconds[$oi];

                if ($p eq $u) { next; }


                # complete
                $s = $origStates[$oi];
                $s = issueCommand("step $s $u");
                while (1) {
                    my @actions2  = split /[\s]+/, issueCommand("actions $s");
                    if (@actions2 == 0) { last; }  # final state
                    my @features2;
                    if ($cfg{'actions'} eq 'fixed') {
                        @features2 = split /[\s]+/, issueCommand("features $s");
                    } else {
                        for (my $i=0; $i<@actions2; $i++) {
                            @{$features2[$i]} = split /[\s]+/, issueCommand("action_features $s $actions2[$i]");
                        }
                    }
                    my $pol = select_policy(0);
                    my $p = getPrediction($pol, ($XV==1)?0:(-1), \@actions2, \@features2, 0);
                    $s = issueCommand("step $s $p");
                }

                # how much did it cost?
                my $uncertainLoss = issueCommand("loss $s");

                $s = $origStates[$oi];
                my @actions = split /[\s]+/, issueCommand("actions $s");
                if (@actions == 0) { last; }  # final state
                my @features;
                if ($cfg{'actions'} eq 'fixed') {
                    @features = split /[\s]+/, issueCommand("features $s");
                } else {
                    for (my $i=0; $i<@actions; $i++) {
                        @{$features[$i]} = split /[\s]+/, issueCommand("action_features $s $actions[$i]");
                    }
                }

                my $min = ($uncertainLoss < $standardLoss) ? $uncertainLoss : $standardLoss;
                my $max = ($uncertainLoss > $standardLoss) ? $uncertainLoss : $standardLoss;
                my $stepDiff = $max - $min; # $numSteps - $oi;
                if ($max-$min > 0) {
                    @actions = ($p, $u);
                    my @costs   = (($standardLoss - $min) / $stepDiff, ($uncertainLoss - $min) / $stepDiff);
                    for (my $xv=0; $xv<@outputs; $xv++) {
                        my $WR = \*{$outputs[$xv]};
                        if (($XV == 1) || (($n % $XV) != $xv)) {
                            makeComment($WR, "uncertainty " . $origUncert[$oi]);
                            makeExample($WR, \@actions, \@costs, \@features, 0);
                        }
                    }
                    if ($XV > 1) {
                        makeComment(\*D, "uncertainty " . $origUncert[$oi]);
                        makeExample(\*D, \@actions, \@costs, \@features, 0);
                    }
                }
            }    
        }

        print STDERR "\n";
    }

    # tell the search to load the data
    print STDERR "creating examples";
    issueCommand("load $trainFile");
    my $num = issueCommand("num");

    my $count = int($num/50); if ($count <= 0) { $count = 1; }

    my $totalPred = 0; my $optimalPred = 0;

    for (my $n=0; $n<$num; $n++) {
        if (($n % $count) == 0) { print STDERR "."; }

        # add dev if required
        if (($cfg{'classifier_usedev'} eq 'true') && ($n == int($num * 0.85))) { 
            for (my $xv=0; $xv<@outputs; $xv++) { my $WR = \*{$outputs[$xv]}; print $WR "DEV\n"; }
            if ($XV > 1) { print D "DEV\n"; }
        }

        my $s = issueCommand("example $n");  # get start state for example n
        while (1) {
            my @actions  = split /[\s]+/, issueCommand("actions $s");
            if (@actions == 0) { last; }  # final state

            my @costs = (); my $minCost = 0;
            if ($cfg{'costs'} eq 'optimal') {
                for (my $i=0; $i<@actions; $i++) {
                    $costs[$i] = issueCommand("cost $s $actions[$i]");
                    if (($i > 0) && ($costs[$i] < $costs[$minCost])) { $minCost = $i; }
                }
            } else { # have to simulate costs
                my $origState = $s;
                for (my $jj=0; $jj<@actions; $jj++) {
                    $s = issueCommand("step $origState $actions[$jj]");
                    while (1) {
                        my @actions2  = split /[\s]+/, issueCommand("actions $s");
                        if (@actions2 == 0) { last; }  # final state
                        my @features2;
                        if ($cfg{'actions'} eq 'fixed') {
                            @features2 = split /[\s]+/, issueCommand("features $s");
                        } else {
                            for (my $i=0; $i<@actions2; $i++) {
                                @{$features2[$i]} = split /[\s]+/, issueCommand("action_features $s $actions2[$i]");
                            }
                        }
                        my $pol = select_policy(1);
                        my $p;
                        if ($policy{$pol}{TYPE} eq 'optimal') {
                            my @costs2 = (); my $minCost2 = 0;
                            for (my $i=0; $i<@actions2; $i++) {
                                $costs2[$i] = issueCommand("cost $s $actions2[$i]");
                                if (($i > 0) && ($costs2[$i] < $costs2[$minCost2])) { $minCost2 = $i; }
                            }

                            my %okAct = (); $okAct{$minCost2} = 1;
                            for (my $i=0; $i<@actions2; $i++) { if ($costs2[$i] <= $costs2[$minCost2]) { $okAct{$i} = 1; } }

#                            print STDERR "[[" . (join ' ', keys %okAct) . "]]\t((" . (join ' ', @actions2) . "))\n";

                            if (scalar keys %okAct == 1) { 
                                $p = $actions2[$minCost2];
                            } else {
                                my @a = keys %okAct;
                                my $a = $a[int(rand() * scalar @a)];
#                                print STDERR " --> $a\n";
                                $p = $actions2[$a];
                            }
                        } else {
                            $p = getPrediction($pol, ($XV==1)?0:(-1), \@actions2, \@features2, 0);
                        }
#                        print STDERR "{{$s --> $p}}\n";
                        $s = issueCommand("step $s $p");
                    }
                    $costs[$jj] = issueCommand("loss $s");
                    if (($jj > 0) && ($costs[$jj] < $costs[$minCost])) { $minCost = $jj; }
                }
                $s = $origState;
                for (my $i=0; $i<@actions; $i++) {
                    $costs[$i] -= $costs[$minCost];
                }
            }

            my @features;

            if ($cfg{'actions'} eq 'fixed') {
                @features = split /[\s]+/, issueCommand("features $s");
            } else {
                for (my $i=0; $i<@actions; $i++) {
                    @{$features[$i]} = split /[\s]+/, issueCommand("action_features $s $actions[$i]");
                }
            }

            for (my $xv=0; $xv<@outputs; $xv++) {
                my $WR = \*{$outputs[$xv]};
                if (($XV == 1) || (($n % $XV) != $xv)) {
                    makeExample($WR, \@actions, \@costs, \@features, 0);
                }
            }
            if ($XV > 1) {
                makeExample(\*D, \@actions, \@costs, \@features, 0);
            }

            my $pol = select_policy(1);
            if ($policy{$pol}{TYPE} eq 'optimal') {
                # how many options have minimum cost?
                my %okAct = (); $okAct{$minCost} = 1;
                for (my $i=0; $i<@actions; $i++) { if ($costs[$i] <= $costs[$minCost]) { $okAct{$i} = 1; } }
                if (scalar keys %okAct == 1) { 
                    $s = issueCommand("step $s $actions[$minCost]");
                } else {
                    my @a = keys %okAct;
                    my $a = $a[int(rand() * scalar @a)];
                    $s = issueCommand("step $s $actions[$a]");
                }
            } else {
                my $p = getPrediction($pol, $n % $XV, \@actions, \@features, 0);
                $s = issueCommand("step $s $p");
                $totalPred++;
                if ($p eq $actions[$minCost]) { $optimalPred++; }
            }
        } 
    }
    if ($iteration > 1) {
        print STDERR " $totalPred predictions (" . ((int($optimalPred/$totalPred*1000))/10) . "% optimal)";
        if ($optimalPred >= $totalPred * 0.95) { print STDERR " !!!WARNING!!!"; }
    }
    print STDERR "\n";

    for (my $i=0; $i<@outputs; $i++) {
        close $outputs[$i] or die;
    }
    if ($XV > 1) { close D or die; }


    #if ($iteration == 2) { die; }

    # now, train a model
    my $TMPMODEL = ".tmp.$tempnumber.model.$iteration";
    my $LASTMODEL = ".tmp.$tempnumber.model." . ($iteration-1);
    print STDERR "training classifiers...";
    for (my $xv=0; $xv<@outputs; $xv++) {
        if (@outputs > 1) { print STDERR "fold $xv..."; }
        if ($cfg{'classifier_pp'} ne '') {
            my $cmd = $cfg{'classifier_pp'};
            if ($cfg{'classifier_bzip'} eq 'true') {
                system("mv $TMPDATA.$xv.bz2 $TMPDATA.$xv.nopp.bz2");
                $cmd =~ s/\$IN/$TMPDATA.$xv.nopp.bz2/g;
                $cmd =~ s/\$OUT/$TMPDATA.$xv.bz2/g;
            } else {
                system("mv $TMPDATA.$xv $TMPDATA.$xv.nopp");
                $cmd =~ s/\$IN/$TMPDATA.$xv.nopp/g;
                $cmd =~ s/\$OUT/$TMPDATA.$xv/g;
            }
            system("$cmd");
            system("rm -rf $TMPDATA.$xv.nopp*");
        }
        
        my $cmd = $cfg{'classifier_train'};
        if ($iteration == 1) { $cmd =~ s/\$LASTMODEL/./g; }

        if ($cfg{'classifier_bzip'} eq 'true') {
            $cmd =~ s/\$DATA/$TMPDATA.$xv.bz2/g;
            $cmd =~ s/\$MODEL/$TMPMODEL.$xv.bz2/g;
            $cmd =~ s/\$LASTMODEL/$LASTMODEL.$xv.bz2/g;
            system($cmd);
#            system("cp $TMPDATA.$xv.bz2 for_arvind/train.$iteration.bz2");
            system("rm $TMPDATA.$xv.bz2");
        } else {
            $cmd =~ s/\$DATA/$TMPDATA.$xv/g;
            $cmd =~ s/\$MODEL/$TMPMODEL.$xv/g;
            $cmd =~ s/\$LASTMODEL/$LASTMODEL.$xv/g;
            system($cmd);
            system("rm $TMPDATA.$xv");
        }
    }
    if ($XV > 1) {
        print STDERR "global fold...";
        if ($cfg{'classifier_pp'} ne '') {
            my $cmd = $cfg{'classifier_pp'};
            if ($cfg{'classifier_bzip'} eq 'true') {
                system("mv $TMPDATA.bz2 $TMPDATA.nopp.bz2");
                $cmd =~ s/\$IN/$TMPDATA.nopp.bz2/g;
                $cmd =~ s/\$OUT/$TMPDATA.bz2/g;
            } else {
                system("mv $TMPDATA $TMPDATA.nopp");
                $cmd =~ s/\$IN/$TMPDATA.nopp/g;
                $cmd =~ s/\$OUT/$TMPDATA/g;
            }
            system("$cmd");
            system("rm -rf $TMPDATA.nopp*");
        }
        
        my $cmd = $cfg{'classifier_train'};
        if ($iteration == 1) { $cmd =~ s/\$LASTMODEL/./g; }
        if ($cfg{'classifier_bzip'} eq 'true') {
            $cmd =~ s/\$DATA/$TMPDATA.bz2/g;
            $cmd =~ s/\$MODEL/$TMPMODEL.bz2/g;
            $cmd =~ s/\$LASTMODEL/$LASTMODEL.bz2/g;
            system($cmd);
            system("rm $TMPDATA.bz2");
        } else {
            $cmd =~ s/\$DATA/$TMPDATA/g;
            $cmd =~ s/\$MODEL/$TMPMODEL/g;
            $cmd =~ s/\$LASTMODEL/$LASTMODEL/g;
            system($cmd);
            system("rm $TMPDATA");
        }
    }
    print STDERR "\n";

    # add it to the policy
    my $beta = $cfg{'beta'};
    if ($beta eq '1/T') { $beta = 1 / ($iteration+1); }
    elsif ($beta eq '1-1/T') { $beta = 1 - 1 / ($iteration+1); }
    elsif ($beta eq '1/sqrt(T)') { $beta = 1 / sqrt($iteration+1); }
    elsif ($beta eq '1-1/sqrt(T)') { $beta = 1 - 1 / sqrt($iteration+1); }
    add_to_policy($iteration, $TMPMODEL, $beta);

    # garbage collect policy
    if ($cfg{'policy_gc'} > 0) { policy_gc(); }


    # load the test data
    print STDERR "predicting...";
#    open OUTPUT, "| bzip2 -9 > for_arvind/test.$iteration.bz2" or die;

    issueCommand("load $testFile");
    $num = issueCommand("num");
    for (my $n=0; $n<$num; $n++) {
        my $s = issueCommand("example $n");  # get start state for example n
        while (1) {
            my @actions  = split /[\s]+/, issueCommand("actions $s");
            if (@actions == 0) { last; }  # final state
            my @features;
            if ($cfg{'actions'} eq 'fixed') {
                @features = split /[\s]+/, issueCommand("features $s");
            } else {
                for (my $i=0; $i<@actions; $i++) {
                    @{$features[$i]} = split /[\s]+/, issueCommand("action_features $s $actions[$i]");
                }
            }
            my $pol = select_policy(0);
            my $p = getPrediction($pol, ($XV==1)?0:(-1), \@actions, \@features, 0);

            my @unknownCosts = (); for (my $i=0; $i<@actions; $i++) { $unknownCosts[$i] = $i; }
#            makeExample(\*OUTPUT, \@actions, \@unknownCosts, \@features);

            $s = issueCommand("step $s $p");
        }
        issueCommand("loss $s");
    }
#    close OUTPUT;
    my $totalLoss = issueCommand("overall_loss");
    print STDERR "loss = $totalLoss\n\n\n";
}

# at the end, kill the search
issueCommand("quit");
waitpid($searchPid, &WNOHANG);

# and close all policies
foreach my $p (keys %policy) {
    if (not ($policy{$p}{TYPE} eq 'classifier')) { next; }
    free_policy($p);
}

# write the learned policy
foreach my $p (sort { $a <=> $b } keys %policy) {
    if (not ($policy{$p}{TYPE} eq 'classifier')) { next; }
    print $policy{$p}{WEIGHT} . "\t" . $policy{$p}{FNAME} . "\n";
}

sub add_to_policy {
    my ($p, $model, $beta) = @_;

    foreach my $n (keys %policy) {
        $policy{$n}{WEIGHT} *= (1 - $beta);
    }

    $policy{$p}{TYPE} = 'classifier';
    $policy{$p}{WEIGHT} = $beta;
    $policy{$p}{FNAME}  = $model;

    my $XV = $cfg{'xv'}; if ($XV eq 'none') { $XV = 1; }

    for (my $xv=0; $xv<$XV; $xv++) {
        local *ThisPredictR;
        local *ThisPredictW;

        my $cmd = $cfg{'classifier_predict'};
        if ($cfg{'classifier_bzip'} eq 'true') {
            $cmd =~ s/\$MODEL/$model.$xv.bz2/g;
        } else {
            $cmd =~ s/\$MODEL/$model.$xv/g;
        }
        my $thisPredictPid = open2(\*ThisPredictR, \*ThisPredictW, $cmd)
            or die ("cannot open bidirectional pipe for executable " . $cmd);

        $policy{$p}{FILES}{$xv}{READER} = *ThisPredictR{IO};
        $policy{$p}{FILES}{$xv}{WRITER} = *ThisPredictW{IO};
        $policy{$p}{FILES}{$xv}{PID}    = $thisPredictPid;
    }

    if ($XV > 1) {
        local *ThisPredictR;
        local *ThisPredictW;

        my $cmd = $cfg{'classifier_predict'};
        if ($cfg{'classifier_bzip'} eq 'true') {
            $cmd =~ s/\$MODEL/$model.bz2/g;
        } else {
            $cmd =~ s/\$MODEL/$model/g;
        }
        my $thisPredictPid = open2(\*ThisPredictR, \*ThisPredictW, $cmd)
            or die ("cannot open bidirectional pipe for executable " . $cmd);

        $policy{$p}{FILES}{-1}{READER} = *ThisPredictR{IO};
        $policy{$p}{FILES}{-1}{WRITER} = *ThisPredictW{IO};
        $policy{$p}{FILES}{-1}{PID}    = $thisPredictPid;
    }
}

sub free_policy {
    my ($p) = @_;
    foreach my $n (keys %{$policy{$p}{FILES}}) {
        close $policy{$p}{FILES}{$n}{READER};
        close $policy{$p}{FILES}{$n}{WRITER};
        waitpid($policy{$p}{FILES}{$n}{PID}, &WNOHANG);
    }
}

sub policy_gc {
    my $killedMass = 0;
    my $remaining = 0;
    foreach my $p (keys %policy) {
        if ($policy{$p}{TYPE} ne 'classifier') { next; }
        if ($policy{$p}{WEIGHT} < $cfg{'policy_gc'}) {
            if ($killedMass == 0) { print STDERR "garbage collecting..."; }
            print STDERR "policy $p (p=" . $policy{$p}{WEIGHT} . ")...";
            $killedMass += $policy{$p}{WEIGHT};
            free_policy($p);
            delete $policy{$p};
        } else {
            $remaining++;
        }
    }

    if ($killedMass > 0) {
        print STDERR "\n";
        foreach my $p (keys %policy) {
            if ($policy{$p}{TYPE} ne 'classifier') { next; }
            $policy{$p}{WEIGHT} += $killedMass/$remaining;
        }
    }
}

sub select_policy {
    my ($optimalOkay) = @_;

    my $total = 0;
    foreach my $n (keys %policy) {
        if ((not $optimalOkay) && ($n == 0)) { next; }
        $total += $policy{$n}{WEIGHT};
    }

    my $r = rand() * $total;
    foreach my $n (keys %policy) {
        if ((not $optimalOkay) && ($n == 0)) { next; }
        $r -= $policy{$n}{WEIGHT};
        if ($r <= 0) { return $n; }
    }

    die "cannot select policy\n";
}

sub getClassId {
    my ($className) = @_;

    if (not exists $classes{$className}) { 
        $classesR{$classesN} = $className;
        $classes{$className} = $classesN++; 
    }
    return $classes{$className};
}

sub getFeatureList {
    my ($f,$i) = @_;

    my @feat = ();
    if ($cfg{'actions'} eq 'fixed') {
        @feat = @{$f};
    } else {
        @feat = @{$f->[$i]};
    }

    my $val = 0;
    for (my $i=0; $i<@feat; $i+=2) {
        $val += $feat[$i+1] * $feat[$i+1];
    }
    $val = 1 / sqrt($val);

    if ($cfg{'classifier_features'} eq 'text') { 
        my $txt = '';
        for (my $i=0; $i<@feat; $i+=2) { $txt .= $feat[$i] . ' ' . ($feat[$i+1]*$val) . ' '; }
        $txt =~ s/ $//;
        return $txt;
    } elsif (($cfg{'classifier_features'} eq 'numeric') || ($cfg{'classifier_features'} eq 'svm')) {
        my %f = ();
        for (my $i=0; $i<@feat; $i+=2) {
            if (not exists $features{$feat[$i]}) { $features{$feat[$i]} = $featuresN++; }
            $f{$features{$feat[$i]}} = $feat[$i+1];
        }
        my $txt = '';
        foreach my $k (sort { $a <=> $b } keys %f) {
            $txt .= $k . ' ' . ($f{$k}*$val) . ' ';
        }
        if ($cfg{'classifier_features'} eq 'svm') { $txt =~ s/ /:1 /g; }
        $txt =~ s/ $//;
        return $txt;
    }
}

sub getPrediction {
    my ($weightId, $weightNum, $actions, $feats, $returnArray) = @_;

    if (not exists $policy{$weightId}{FILES}{$weightNum}) {
        die "policy does not exist: $weightId $weightNum\n";
    }

    my $WR = \*{$policy{$weightId}{FILES}{$weightNum}{WRITER}};
    my $RE = \*{$policy{$weightId}{FILES}{$weightNum}{READER}};

    my @costs = ();
    for (my $i=0; $i<@{$actions}; $i++) { $costs[$i] = $i; }


#    print STDERR "prediction example\n";
    makeExample($WR, $actions, \@costs, $feats, 1);

    if (($cfg{'classifier_type'} eq 'multiclass') || ($cfg{'classifier_type'} eq 'multilabel')) {
        my $result = <$RE>;
        chomp $result;
#        print STDERR "result = $result\n";
        my ($classPred, @probs) = split /[\s]+/, $result;

        my $cl = -1; my $v = 0;
        for (my $i=0; $i<@{$actions}; $i++) {
            my $id = getClassId($actions->[$i]);
            if ($id >= @probs) { next; }
            if (($cl < 0) || ($probs[$id] > $v)) {
                $cl = $i;
                $v  = $probs[$id];
            }
        }
        if ($returnArray) {
            return ($actions->[$cl], @probs);
        } else {
            return $actions->[$cl];
        }
    } elsif ($cfg{'classifier_type'} eq 'singletons') {
        my @probs = ();
        for (my $i=0; $i<@{$actions}; $i++) { $probs[$i] = 0; }

        for (my $i=0; $i<@{$actions}; $i++) {
            my $result = <$RE>;
            chomp $result;
            my ($c,$p) = split /[\s]+/, $result;
            $probs[$i] = $p;
        }

        my $cl = -1;
        for (my $i=0; $i<@{$actions}; $i++) {
            if ($i >= @probs) { next; }
            if (($cl < 0) || ($probs[$i] > $probs[$cl])) {
                $cl = $i;
            }
        }
        if ($returnArray) {
            return ($actions->[$cl], @probs);
        } else {
            return $actions->[$cl];
        }
    }
}

sub makeComment {
    my ($H, $comment) = @_;

    print $H "# $comment\n";
}

sub makeExample {
    my ($H, $actions, $costs, $f, $forceMake) = @_;

    my $minCost = 0;
    my $maxCost = 0;
    for (my $i=1; $i<@{$costs}; $i++) {
        if ($costs->[$i] < $costs->[$minCost]) { $minCost = $i; }
        if ($costs->[$i] > $costs->[$maxCost]) { $maxCost = $i; }
    }

#    print STDERR $cfg{'classifier_type'} . "\t" . $costs->[$minCost] . "\t" . $costs->[$maxCost] . "\n";

    if (($costs->[$minCost] >= $costs->[$maxCost]) && (not $forceMake)) { 
        #all costs equal, not forced
        return; 
    }
#    print STDERR "{$costs->[$minCost] < $costs->[$maxCost]}";

    if ($cfg{'classifier_type'} eq 'multiclass') {
        print $H getClassId($actions->[$minCost]) . ' ' . getFeatureList($f) . "\n";
    } elsif ($cfg{'classifier_type'} eq 'multilabel') {

#        print STDERR "makeExample:\n";
        my @carr = ();
        for (my $c=0; $c<$classesN; $c++) { $carr[$c] = 'x'; } # $costs->[$maxCost]; }  # bad class
        for (my $i=0; $i<@{$actions}; $i++) {
            my $v = $costs->[$i] - $costs->[$minCost];
            $carr[getClassId($actions->[$i])] = $v;
        }

#        print STDERR ((join ':', @carr) . ' ' . getFeatureList($f) . "\n");
        print $H ((join ':', @carr) . ' ' . getFeatureList($f) . "\n");
    } elsif ($cfg{'classifier_type'} eq 'singletons') {
        for (my $i=0; $i<@{$actions}; $i++) {
            print $H (($costs->[$i] <= $costs->[$minCost])?"1":"-1") . ' ' . getFeatureList($f,$i) . "\n";
        }
    }
}

sub issueCommand {
    my ($cmd) = @_;

    print SearchW "$cmd\n";
    my $result = <SearchR>;
    chomp $result;
    if ($result =~ /^OK(.*)$/) {
        my $r = $1; $r =~ s/^[\s]*//; $r =~ s/[\s]*$//;
        return $r;
    } else {
        die "search script did not return OK for comment: '$cmd'\n";
    }
}

sub load_config {
    my ($fp) = @_;

    my %cfg = ();

    $cfg{'name'}  = '';

    $cfg{'beta'}  = '';
    $cfg{'costs'} = '';
    $cfg{'xv'}    = '';
    $cfg{'policy_gc'} = '';
    $cfg{'actions'} = '';
    $cfg{'active_iteration'} = 999999;
    $cfg{'active_data'} = '';

    $cfg{'search_exec'} = '';

    $cfg{'classifier_bzip'} = 'false';
    $cfg{'classifier_type'} = '';
    $cfg{'classifier_usedev'} = '';
    $cfg{'classifier_features'} = '';
    $cfg{'classifier_train'} = '';
    $cfg{'classifier_predict'} = '';
    $cfg{'classifier_pp'} = '';
    $cfg{'classifier_predict_pp'} = '';

    open F, $fp or die "could not open $fp for config info\n";
    while (<F>) {
        chomp;
        s/\#.*$//;
        s/^[\s]+/ /g; s/^ //; s/ $//;
        if (/^[\s]*$/) { next; }
        if (/^([^:]+)[\s]*:[\s]*(.+)[\s]*$/) {
            my $k = $1; my $v = $2;
            if (not exists $cfg{$k}) {
                die "property '$k' unknown in config $fp\n";
            }
            $v =~ s/ *$//;
            $cfg{$k} = $v;
        }
    }
    close F or die;

    foreach my $k (keys %cfg) {
        my $v = $cfg{$k};
        if (not exists $ALLOWEDVALUES{$k}) { next; }
        if (not exists $ALLOWEDVALUES{$k}{$v}) {
            if ($v eq '') { die "property '$k' must be set in config $fp\n"; }
            die "property '$k' cannot take value '$v' in config $fp\n";
        }
    }
    

    return (%cfg);
}

sub uncertainty {
    my @p = @_;
    my $u = 0;
    my $best = 0;
    for (my $i=0; $i<@p; $i++) {
        if (($p[$i] > 0) && ($p[$i] < 1)) { $u -= $p[$i] * log($p[$i]); }
        if ($p[$i] > $p[$best]) { $best = $i; }
    }
    my $snd = 0; if ($best == 0) { $snd = 1; }
    for (my $i=0; $i<@p; $i++) {
        if ($i == $best) { next; }
        if ($p[$i] > $p[$snd]) { $snd = $i; }
    }
#    print STDERR "($p[$best] - $p[$snd])";
    #$u = 1 / ($p[$best] - $p[$snd]);
    return ($u,$snd);
}

sub random_step {
    my @p = @_;
    my $r = rand;
    for (my $i=0; $i<@p; $i++) {
        $r -= $p[$i];
        if ($r <= 0) { return $i; }
    }
    return 0;
}

sub REAPER {
    my $stiff;
    while (($stiff = waitpid(-1, &WNOHANG)) > 0) {
        # do something with $stiff if you want
    }
    $SIG{CHLD} = \&REAPER;                  # install *after* calling waitpid
}

